import numpy as np
from fractions import Fraction
from math import isclose
from poisson_approval.best_response.BestResponse import BestResponse
from poisson_approval.constants.constants import *
from poisson_approval.utils.UtilCache import cached_property


class BestResponseApproval(BestResponse):
    """Best response for a given ordinal type of voter in Approval voting.

    The main objective of this class is to compute :attr:`threshold_utility`.

    It also provides the string :attr:`justification`, indicating which sub-algorithm was used. Nowadays, possible
    values are ``'Asymptotic method'``, ``'Simplified asymptotic method'``, ``'Easy vs difficult pivot'``,
    ``'Difficult vs easy pivot'``, ``'Offset method'``, ``'Offset method with trio approximation correction'``.

    Parameters
    ----------
    Cf. :class:`BestResponse`.

    Attributes
    ----------
    Cf. :class:`BestResponse`
    """

    ASYMPTOTIC = 'Asymptotic method'
    ASYMPTOTIC_SIMPLIFIED = 'Simplified asymptotic method'
    EASY_VS_DIFFICULT = 'Easy vs difficult pivot'
    DIFFICULT_VS_EASY = 'Difficult vs easy pivot'
    OFFSET_METHOD = 'Offset method'
    OFFSET_METHOD_WITH_TRIO_APPROXIMATION_CORRECTION = 'Offset method with trio approximation correction'

    voting_rule = APPROVAL

    # =======
    # Results
    # =======

    @cached_property
    def results_asymptotic_method(self):
        """tuple (threshold_utility, justification) : Results according to the asymptotic method. Cf.
        :attr:`threshold_utility` and :attr:`justification`. The threshold utility may be NaN, because this method is
        not always sufficient.
        """
        threshold_utility = ((
            self.pivot_tij.asymptotic * Fraction(1, 2)
            + self.trio_1t.asymptotic * Fraction(1, 3) + self.trio_2t.asymptotic * Fraction(1, 6)
        ) / (
            self.pivot_tij.asymptotic * Fraction(1, 2) + self.pivot_tjk.asymptotic * Fraction(1, 2)
            + self.trio_1t.asymptotic * Fraction(2, 3) + self.trio_2t.asymptotic * Fraction(1, 3)
        )).limit
        justification = self.ASYMPTOTIC
        return threshold_utility, justification

    @cached_property
    def results_limit_pivot_theorem(self):
        """tuple (threshold_utility, justification) : Results according to the limit pivot theorem.
        Cf. :attr:`threshold_utility` and :attr:`justification`. If the tau-vector has two consecutive zeros, the
        theorem does not apply and this method returns ``nan, ''``.
        """
        if self.tau.has_two_consecutive_zeros:
            return np.nan, ''
        if self.pivot_ij_easy_or_tight and self.pivot_jk_easy_or_tight:
            # Both pivots are easy => We can forget the trios.
            threshold_utility = ((
                self.pivot_tij.asymptotic * Fraction(1, 2)
            ) / (
                self.pivot_tij.asymptotic * Fraction(1, 2) + self.pivot_tjk.asymptotic * Fraction(1, 2)
            )).limit
            justification = self.ASYMPTOTIC_SIMPLIFIED
        elif self.pivot_ij_easy_or_tight:
            # ... but pivot jk is difficult.
            threshold_utility = 1
            justification = self.EASY_VS_DIFFICULT
        elif self.pivot_jk_easy_or_tight:
            # ... but pivot ij is difficult.
            threshold_utility = 0
            justification = self.DIFFICULT_VS_EASY
        else:
            # Both pivots are difficult => offset method.
            # Due to approximations in trio event, psi_k and psi_i may exceptionally be greater than 1 (whereas
            # in difficult pivots, we know that they must be strictly lower than 1). In that case, the formulas
            # for the offset method will fail, so we must be cautious.
            psi_k_greater_but_close_to_one = False
            if self.trio.psi[self.k] >= 1:
                if isclose(self.trio.psi[self.k], 1, rel_tol=1e-1):
                    psi_k_greater_but_close_to_one = True
                else:  # pragma: no cover
                    raise AssertionError('Unexpected: self.trio.psi[self.k] = %s > 1' % self.trio.psi[self.k])
            psi_i_greater_but_close_to_one = False
            if self.trio.psi[self.i] >= 1:
                if isclose(self.trio.psi[self.i], 1, rel_tol=1e-1):
                    psi_i_greater_but_close_to_one = True
                else:  # pragma: no cover
                    raise AssertionError('Unexpected: self.trio.psi[self.i] = %s > 1' % self.trio.psi[self.i])
            if psi_i_greater_but_close_to_one and psi_k_greater_but_close_to_one:  # pragma: no cover
                raise AssertionError('Unexpected: both psi_i and psi_k are greater and close to 1.')
            elif psi_k_greater_but_close_to_one:
                # pij ~= inf, pjk < inf ==> u = 1
                threshold_utility = 1
                justification = self.OFFSET_METHOD_WITH_TRIO_APPROXIMATION_CORRECTION
            elif psi_i_greater_but_close_to_one:
                # pij < inf, pjk ~= inf ==> u = 0
                threshold_utility = 0
                justification = self.OFFSET_METHOD_WITH_TRIO_APPROXIMATION_CORRECTION
            else:
                # General case of the offset method (at last!)
                pij = (1 + self.trio.psi[self.ik]) / (1 - self.trio.psi[self.k])
                pjk = (1 + self.trio.psi[self.j]) * self.trio.psi[self.i] ** 2 / (1 - self.trio.psi[self.i])
                p1t = self.trio.psi[self.i]
                p2t = self.trio.psi[self.ij]
                threshold_utility = (pij / 2 + p1t / 3 + p2t / 6) / (pij / 2 + pjk / 2 + p1t * 2 / 3 + p2t / 3)
                justification = self.OFFSET_METHOD
        return threshold_utility, justification

    @cached_property
    def results(self):
        """tuple (threshold_utility, justification) : Cf. :attr:`threshold_utility` and :attr:`justification`.
        These results use:

        * :meth:`results_asymptotic_method` if there are two consecutive zeros in the "compass diagram" of the
          tau-vector,
        * :meth:`results_limit_pivot_theorem` otherwise.
        """
        if self.tau.has_two_consecutive_zeros:
            return self.results_asymptotic_method
        else:
            return self.results_limit_pivot_theorem



from poisson_approval.utils.Util import sort_ballot, ballot_one, ballot_two, ballot_one_two, ballot_one_three
from poisson_approval.constants.constants import *
from poisson_approval.strategies.Strategy import Strategy
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros


# noinspection PyUnresolvedReferences
class StrategyTwelve(Strategy):
    """A strategy profile for a 12-type profile.

    Parameters
    ----------
    d_ranking_ballot : dict
        Keys are rankings and values are strategies, e.g. ``'abc': 'ab'``. A strategy can be a valid ballot
        (e.g. ``'a'`` or ``'ab'`` when the ranking is ``'abc'``), ``UTILITY_DEPENDENT`` or ``''`` if the behavior
        of these voters is not specified in the strategy. Cf. :class:`ProfileTwelve`.

        In the case of Plurality, the valid ballots for voters ``'abc'`` are ``'a'`` and ``'b'``. In the case of
        Anti-plurality, their valid ballots are ``'ab'`` (vote against `c`) and ``'ac'`` (vote against `b`).
    profile : Profile, optional
        The "context" in which the strategy is used.
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``. Default: the same
        voting rule as `profile` if a profile is specified, ``APPROVAL`` otherwise.

    Examples
    --------
        >>> strategy = StrategyTwelve({'abc': 'utility-dependent', 'bac': 'b'})
        >>> strategy
        StrategyTwelve({'abc': 'utility-dependent', 'bac': 'b'})
        >>> print(strategy)
        <abc: utility-dependent, bac: b>
        >>> strategy.abc
        'utility-dependent'
        >>> strategy.a_bc
        'a'
        >>> strategy.ab_c
        'ab'
    """

    def __init__(self, d_ranking_ballot, profile=None, voting_rule=None):
        """
            >>> strategy = StrategyTwelve({'non_existing_ranking': 'utility-dependent'})
            Traceback (most recent call last):
            ValueError: Unknown key: non_existing_ranking
            >>> strategy = StrategyTwelve({'abc': 'non_existing_ballot'})
            Traceback (most recent call last):
            ValueError: Unknown strategy: non_existing_ballot
        """
        voting_rule = self._get_voting_rule_(profile, voting_rule)
        # Populate the dictionary and check for typos in the input
        self.d_ranking_ballot = DictPrintingInOrderIgnoringZeros({ranking: '' for ranking in RANKINGS})
        for ranking, ballot in d_ranking_ballot.items():
            # Sanity checks
            if ranking not in RANKINGS:
                raise ValueError('Unknown key: ' + ranking)
            if voting_rule == APPROVAL:
                authorized_ballots = {ballot_one(ranking), ballot_one_two(ranking), ballot_one_two(ranking)[::-1],
                                      '', UTILITY_DEPENDENT}
            elif voting_rule == PLURALITY:
                authorized_ballots = {ballot_one(ranking), ballot_two(ranking), '', UTILITY_DEPENDENT}
            elif voting_rule == ANTI_PLURALITY:
                authorized_ballots = {ballot_one_two(ranking), ballot_one_two(ranking)[::-1],
                                      ballot_one_three(ranking), ballot_one_three(ranking)[::-1],
                                      '', UTILITY_DEPENDENT}
            else:
                raise NotImplementedError
            if ballot not in authorized_ballots:
                raise ValueError('Unknown strategy: ' + ballot)
            # Record the ballot
            self.d_ranking_ballot[ranking] = ballot if ballot == UTILITY_DEPENDENT else sort_ballot(ballot)
        # Call parent class
        super().__init__(profile=profile, voting_rule=voting_rule)

    def __eq__(self, other):
        """Equality test.

        Parameters
        ----------
        other : object

        Returns
        -------
        bool
            True if this strategy is equal to `other`.

        Examples
        --------
            >>> strategy = StrategyTwelve({'abc': 'utility-dependent', 'bac': 'b'})
            >>> strategy == StrategyTwelve({'abc': 'utility-dependent', 'bac': 'b'})
            True
        """
        return (isinstance(other, StrategyTwelve)
                and self.d_ranking_ballot == other.d_ranking_ballot
                and self.voting_rule == other.voting_rule)

    # Representation

    def __repr__(self):
        arguments = repr(self.d_ranking_ballot)
        if self.voting_rule != APPROVAL:
            arguments += ', voting_rule=%r' % self.voting_rule
        return 'StrategyTwelve(%s)' % arguments

    def __str__(self):
        result = '<%s>' % str(self.d_ranking_ballot)[1:-1]
        if self.profile is not None:
            result += ' ==> ' + str(self.winners)
        if self.voting_rule != APPROVAL:
            result += ' (%s)' % self.voting_rule
        return result

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')


def make_property_ranking_ballot(ranking, doc):
    def _f(self):
        return self.d_ranking_ballot[ranking]
    _f.__doc__ = doc
    return property(_f)


for my_ranking in RANKINGS:
    setattr(StrategyTwelve, my_ranking, make_property_ranking_ballot(my_ranking,
                                                                     'str : Ballot of ranking ``%s``.' % my_ranking))


def make_property_ranking_low_u_ballot(ranking, doc):
    def _f(self):
        if getattr(self, ranking) == UTILITY_DEPENDENT:
            if getattr(self, 'voting_rule') in {APPROVAL, PLURALITY}:
                return ballot_one(ranking)
            elif getattr(self, 'voting_rule') == ANTI_PLURALITY:
                return ballot_one_three(ranking)
            else:
                raise NotImplementedError
        return getattr(self, ranking)
    _f.__doc__ = doc
    return property(_f)


for my_ranking in RANKINGS:
    my_t = my_ranking[0] + '_' + my_ranking[1:]
    setattr(StrategyTwelve, my_t, make_property_ranking_low_u_ballot(my_ranking,
                                                                     'str : Ballot of type ``%s``.' % my_t))


def make_property_ranking_high_u_ballot(ranking, doc):
    def _f(self):
        if getattr(self, ranking) == UTILITY_DEPENDENT:
            if getattr(self, 'voting_rule') in {APPROVAL, ANTI_PLURALITY}:
                return ballot_one_two(ranking)
            elif getattr(self, 'voting_rule') == PLURALITY:
                return ballot_two(ranking)
            else:
                raise NotImplementedError
        return getattr(self, ranking)
    _f.__doc__ = doc
    return property(_f)


for my_ranking in RANKINGS:
    my_t = my_ranking[:2] + '_' + my_ranking[2]
    setattr(StrategyTwelve, my_t, make_property_ranking_high_u_ballot(my_ranking,
                                                                      'str : Ballot of type ``%s``.' % my_t))



class AnalyzedStrategies:
    """A container with all the strategies analyzed for a given profile.

    Parameters
    ----------
    equilibria : list of :class:`Strategy`
        List of the strategies where equilibrium is sure.
    utility_dependent : list of :class:`Strategy`
        List of the strategies where the equilibrium depends on the exact utilities.
    inconclusive : list of :class:`Strategy`
        List of the strategies where we are not able to decide whether it is an equilibrium or not.
    non_equilibria : list of :class:`Strategy`
        List of the strategies where the program certifies there is no equilibrium.
    """

    def __init__(self, equilibria: list, utility_dependent: list, inconclusive: list, non_equilibria: list):
        self.equilibria = equilibria
        self.utility_dependent = utility_dependent
        self.inconclusive = inconclusive
        self.non_equilibria = non_equilibria

    def __repr__(self):
        """
        Examples
        --------
        With plurals ("equilibria"):

            >>> from poisson_approval import StrategyOrdinal, ProfileOrdinal
            >>> profile = ProfileOrdinal({'abc': 0.2, 'acb': 0.3, 'bac': 0.5})
            >>> analyzed_strategies = AnalyzedStrategies(
            ...     equilibria=[StrategyOrdinal({'abc': 'a', 'acb': 'a', 'bac': 'b'}, profile),
            ...                 StrategyOrdinal({'abc': 'a', 'acb': 'a', 'bac': 'ab'}, profile)],
            ...     utility_dependent=[StrategyOrdinal({'abc': 'a', 'acb': 'ac', 'bac': 'b'}, profile),
            ...                        StrategyOrdinal({'abc': 'a', 'acb': 'ac', 'bac': 'ab'}, profile)],
            ...     non_equilibria=[StrategyOrdinal({'abc': 'ab', 'acb': 'a', 'bac': 'b'}, profile),
            ...                     StrategyOrdinal({'abc': 'ab', 'acb': 'a', 'bac': 'ab'}, profile)],
            ...     inconclusive=[StrategyOrdinal({'abc': 'ab', 'acb': 'ac', 'bac': 'b'}, profile),
            ...                   StrategyOrdinal({'abc': 'ab', 'acb': 'ac', 'bac': 'ab'}, profile)]
            ... )
            >>> analyzed_strategies
            Equilibria:
            <abc: a, acb: a, bac: b> ==> a, b (FF)
            <abc: a, acb: a, bac: ab> ==> a (FF)
            <BLANKLINE>
            Utility-dependent equilibria:
            <abc: a, acb: ac, bac: b> ==> a, b (FF)
            <abc: a, acb: ac, bac: ab> ==> a (D)
            <BLANKLINE>
            Non-equilibria:
            <abc: ab, acb: a, bac: b> ==> b (FF)
            <abc: ab, acb: a, bac: ab> ==> a (FF)
            <BLANKLINE>
            Inconclusive strategy profiles:
            <abc: ab, acb: ac, bac: b> ==> b
            <abc: ab, acb: ac, bac: ab> ==> a

        With singulars ("equilibrium"):

            >>> analyzed_strategies = AnalyzedStrategies(
            ...     equilibria=[StrategyOrdinal({'abc': 'a', 'acb': 'a', 'bac': 'b'}, profile)],
            ...     utility_dependent=[StrategyOrdinal({'abc': 'a', 'acb': 'ac', 'bac': 'b'}, profile)],
            ...     non_equilibria=[StrategyOrdinal({'abc': 'ab', 'acb': 'a', 'bac': 'b'}, profile)],
            ...     inconclusive=[StrategyOrdinal({'abc': 'ab', 'acb': 'ac', 'bac': 'b'}, profile)]
            ... )
            >>> analyzed_strategies
            Equilibrium:
            <abc: a, acb: a, bac: b> ==> a, b (FF)
            <BLANKLINE>
            Utility-dependent equilibrium:
            <abc: a, acb: ac, bac: b> ==> a, b (FF)
            <BLANKLINE>
            Non-equilibrium:
            <abc: ab, acb: a, bac: b> ==> b (FF)
            <BLANKLINE>
            Inconclusive strategy profile:
            <abc: ab, acb: ac, bac: b> ==> b

        With empty lists, equilibria and non-equilibria still appear, but utility-dependent and inconclusive are not
        even mentioned:

            >>> analyzed_strategies = AnalyzedStrategies(
            ...     equilibria=[],
            ...     utility_dependent=[],
            ...     non_equilibria=[],
            ...     inconclusive=[]
            ... )
            >>> analyzed_strategies
            Equilibria:
            None
            <BLANKLINE>
            Non-equilibria:
            None
        """
        # Equilibria (print even if there is none).
        if len(self.equilibria) == 1:
            s = 'Equilibrium:'
        else:
            s = 'Equilibria:'
        if not self.equilibria:
            s += '\nNone'
        for strategy in self.equilibria:
            s += '\n' + str(strategy) + ' (' + str(strategy.tau.focus) + ')'
        # Print utility-dependent strategies only if there are some (this restriction is useful for ProfileTwelve).
        if len(self.utility_dependent) == 1:
            s += '\n\nUtility-dependent equilibrium:'
        elif len(self.utility_dependent) > 1:
            s += '\n\nUtility-dependent equilibria:'
        for strategy in self.utility_dependent:
            s += '\n' + str(strategy) + ' (' + str(strategy.tau.focus) + ')'
        # Non-equilibria (print even if there is none).
        if len(self.non_equilibria) == 1:
            s += '\n\nNon-equilibrium:'
        else:
            s += '\n\nNon-equilibria:'
        if not self.non_equilibria:
            s += '\nNone'
        for strategy in self.non_equilibria:
            s += '\n' + str(strategy) + ' (' + str(strategy.tau.focus) + ')'
        # Print inconclusive strategies only if there are some (which should not happen).
        if len(self.inconclusive) == 1:
            s += '\n\nInconclusive strategy profile:'
        elif len(self.inconclusive) > 1:
            s += '\n\nInconclusive strategy profiles:'
        for strategy in self.inconclusive:
            s += '\n' + str(strategy)
        return s
